/*
 * Copyright (c) 2006-2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <portLib.h>

#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fcntl.h>
#include <linux/input.h>

#include "server/mouse3dHeader.h"


/* --- local data ---------------------------------------------------------- */

static void		m3d_led(int fd, int on);
static int		m3d_open(const char *path);
static const char *	m3d_finddev(void);

static int mousefd = -1;

/*------------------------------------------------------------------------
 *
 * mouse3dMainInit  --  Initialization codel (fIDS, ...)
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
mouse3dMainInit(int *report)
{
  SDI_F->connected = 0;
  mousefd = -1;

  /* default bindings */
  SDI_F->bindings.x = ABS_X;
  SDI_F->bindings.y = ABS_Y;
  SDI_F->bindings.z = ABS_Z;
  SDI_F->bindings.rx = ABS_RX;
  SDI_F->bindings.ry = ABS_RY;
  SDI_F->bindings.rz = ABS_RZ;
  SDI_F->bindings.b0 = BTN_0;
  SDI_F->bindings.b1 = BTN_1;

  /* set initial readings to 0 */
  SDI_F->measures.x = SDI_F->measures.y = SDI_F->measures.z = 0.;
  SDI_F->measures.Rx = SDI_F->measures.Ry = SDI_F->measures.Rz = 0.;
  SDI_F->measures.b0 = SDI_F->measures.b1 = 0;
  mouse3dMeasuresPosterWrite(MOUSE3D_MEASURES_POSTER_ID, &SDI_F->measures);

  /* set gains to indentity */
  SDI_F->gains.x = SDI_F->gains.y = SDI_F->gains.z = 1.;
  SDI_F->gains.rx = SDI_F->gains.ry = SDI_F->gains.rz = 1.;

  /* set thresholds to 0 */
  SDI_F->thresholds.x = SDI_F->thresholds.y = SDI_F->thresholds.z = 0.;
  SDI_F->thresholds.rx = SDI_F->thresholds.ry = SDI_F->thresholds.rz = 0.;

  return OK;
}

/*------------------------------------------------------------------------
 *
 * mouse3dMainEnd  --  Termination codel
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
mouse3dMainEnd(void)
{
  SDI_F->connected = 0;
  if (mousefd >= 0) {
    m3d_led(mousefd, 0);
    close(mousefd);
  }
  mousefd = -1;
  return OK;
}

/* --- mouse3dExport ------------------------------------------------------- */

ACTIVITY_EVENT
mouse3dExportMain(int *report)
{
  struct input_event ev;
  ssize_t s;
  int e;

  if (mousefd < 0) {
    *report = S_mouse3d_DEVICE_NOT_CONNECTED;
    return ETHER;
  }

  do {
    struct timeval timeout = { 0, 500000 /* 500ms */ };
    fd_set rfd;

    /* wait for event, at most 500ms to let other requests execute */
    FD_ZERO(&rfd);
    FD_SET(mousefd, &rfd);
    e = select(mousefd+1, &rfd, NULL, NULL, &timeout);
    /* if no event, return */
    if (e < 1) break;

    /* read next event */
    do {
      s = read(mousefd, &ev, sizeof(ev));
    } while(s < 0 && errno == EINTR);

    if (s < 0 && errno == EAGAIN) continue;
    if (s != sizeof(ev)) {
      warnx("short read %d", s);
      close(mousefd);
      mousefd = -1;
      break;
    }

    switch(ev.type) {
      case EV_REL:
	warnx("relative events not handled");
	break;

      case EV_ABS:
	if (ev.code == SDI_F->bindings.x) {
	  SDI_F->measures.x = ev.value * SDI_F->gains.x;
	  if (fabs(SDI_F->measures.x) < SDI_F->thresholds.x)
	    SDI_F->measures.x = 0.;
	} else if (ev.code == SDI_F->bindings.y) {
	  SDI_F->measures.y = ev.value * SDI_F->gains.y;
	  if (fabs(SDI_F->measures.y) < SDI_F->thresholds.y)
	    SDI_F->measures.y = 0.;
	} else if (ev.code == SDI_F->bindings.z) {
	  SDI_F->measures.z = ev.value * SDI_F->gains.z;
	  if (fabs(SDI_F->measures.z) < SDI_F->thresholds.z)
	    SDI_F->measures.z = 0.;
	} else if (ev.code == SDI_F->bindings.rx) {
	  SDI_F->measures.Rx = ev.value * SDI_F->gains.rx;
	  if (fabs(SDI_F->measures.Rx) < SDI_F->thresholds.rx)
	    SDI_F->measures.Rx = 0.;
	} else if (ev.code == SDI_F->bindings.ry) {
	  SDI_F->measures.Ry = ev.value * SDI_F->gains.ry;
	  if (fabs(SDI_F->measures.Ry) < SDI_F->thresholds.ry)
	    SDI_F->measures.Ry = 0.;
	} else if (ev.code == SDI_F->bindings.rz) {
	  SDI_F->measures.Rz = ev.value * SDI_F->gains.rz;
	  if (fabs(SDI_F->measures.Rz) < SDI_F->thresholds.rz)
	    SDI_F->measures.Rz = 0.;
	}
	break;

      case EV_KEY:
	/* switch button state on button release event */
	if (!ev.value) {
	  if (ev.code == SDI_F->bindings.b0)
	    SDI_F->measures.b0 = !SDI_F->measures.b0;
	  else if (ev.code == SDI_F->bindings.b1)
	    SDI_F->measures.b1 = !SDI_F->measures.b1;
	}
	break;

      case EV_SYN:
	mouse3dMeasuresPosterWrite(MOUSE3D_MEASURES_POSTER_ID, &SDI_F->measures);
	break;
    }
  } while(1);

  return EXEC;
}

ACTIVITY_EVENT
mouse3dExportEnd(int *report)
{
  SDI_F->measures.x = SDI_F->measures.y = SDI_F->measures.z = 0.;
  SDI_F->measures.Rx = SDI_F->measures.Ry = SDI_F->measures.Rz = 0.;
  SDI_F->measures.b0 = SDI_F->measures.b1 = 0;
  mouse3dMeasuresPosterWrite(MOUSE3D_MEASURES_POSTER_ID, &SDI_F->measures);
  return ETHER;
}


/*------------------------------------------------------------------------
 * Init
 *
 * Description:
 *
 * Reports:      OK
 *              S_mouse3d_DEVICE_NOT_FOUND
 */

/* mouse3dInitMain  -  codel EXEC of Init
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
mouse3dInitMain(mouse3d_path *device, int *report)
{
  const char *path = NULL;

  if (mousefd >= 0) {
    m3d_led(mousefd, 0);
    close(mousefd);
    mousefd = -1;
  }

  path = device->path[0] ? device->path : m3d_finddev();
  if (!path) {
    warnx("failed to detect to the space navigator mouse");
    *report = S_mouse3d_DEVICE_NOT_FOUND;
    return ETHER;
  }

  mousefd = m3d_open(path);
  if (mousefd < 0) {
    *report = S_mouse3d_DEVICE_NOT_FOUND;
    return ETHER;
  }

  m3d_led(mousefd, 1);
  SDI_F->connected = 1;
  warnx("space navigator mouse initialized");
  return ETHER;
}


/* --- m3d_led ------------------------------------------------------------- */

static void
m3d_led(int fd, int on)
{
  struct input_event ev = { 0 };

  ev.type = EV_LED;
  ev.code = LED_MISC;
  ev.value = on;

  (void)write(fd, &ev, sizeof(ev));
}


/* --- m3d_open ------------------------------------------------------------ */

static int
m3d_open(const char *path)
{
  int fd, fl;

  fd = open(path, O_RDWR);
  if (fd < 0) {
    fd = open(path, O_RDONLY);
    if (fd < 0) {
      warn("failed to open %s", path);
      return -1;
    }

    warnx("opened %s read-only, LEDs won't work", path);
  }

  /* fl = fcntl(fd, F_GETFL); */
  /* fl = fcntl(fd, F_SETFL, fl | O_NONBLOCK); */
  /* if (fl < 0) { */
  /*   close (fd); */
  /*   warn("cannot set %s non-blocking", path); */
  /*   return -1; */
  /* } */

  return fd;
}


/* --- m3d_finddev --------------------------------------------------------- */

static const char *
m3d_finddev()
{
  static mouse3d_path device;
  unsigned int i;
  char name[64];
  int fd, s;

  for(i=0; ; i++) {
    snprintf(device.path, sizeof(device.path), "/dev/input/event%d", i);

    fd = open(device.path, O_RDONLY);
    if (fd < 0) {
      if (errno == ENOENT) break;
      warn("failed to open %s", device.path);
      continue;
    }

    s = ioctl(fd, EVIOCGNAME(sizeof(name)), name);
    close(fd);

    if (s == -1) {
      warnx("cannot get name of %s", device.path);
      continue;
    }

    warnx("%s: %s", device.path, name[0] ? name : "unknown");

    if (strstr(name, "3Dconnexion"))
      return device.path;
  }

  return NULL;
}
